# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import eventlet
eventlet.monkey_patch()
import time

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging

from neutron.agent.common import config as a_config
from neutron.common import config as n_config
from pypowervm.tasks import network_bridger as net_br

from networking_powervm._i18n import _LE
from networking_powervm._i18n import _LI
from networking_powervm._i18n import _LW
from networking_powervm.plugins.ibm.agent.powervm import agent_base
from networking_powervm.plugins.ibm.agent.powervm import constants as p_const
from networking_powervm.plugins.ibm.agent.powervm import utils

import sys


LOG = logging.getLogger(__name__)


agent_opts = [
    cfg.StrOpt('bridge_mappings',
               default='',
               help='The Network Bridge mappings (defined by the SEA) that '
                    'describe how the neutron physical networks map to the '
                    'Shared Ethernet Adapters.'
                    'Format: <ph_net1>:<sea1>:<vio1>,<ph_net2>:<sea2>:<vio2> '
                    'Example: default:ent5:vios_1,speedy:ent6:vios_1'),
    cfg.IntOpt('pvid_update_loops', default=180,
               help='The Port VLAN ID (PVID) of the Client VM\'s Network '
                    'Interface is updated by this agent.  There is a delay '
                    'from Nova between when the Neutron Port is assigned '
                    'to the host, and when the client VIF is created.  This '
                    'variable indicates how many loops the agent should take '
                    'until it determines that the port has failed to create '
                    'from Nova.  If no requests are in the system, the loop '
                    'will wait a second before checking again.  If requests '
                    'are in the system, it may take a bit longer.'),
    cfg.BoolOpt('automated_powervm_vlan_cleanup', default=True,
                help='Determines whether or not the VLANs will be removed '
                     'from the Network Bridge if a VM is removed and it is '
                     'the last VM on the system to use that VLAN.  By '
                     'default, will clean up VLANs to improve the overall '
                     'system performance (by reducing broadcast domain).  '
                     'Will only apply to VLANs not on the primary PowerVM '
                     'virtual Ethernet adapter of the SEA.')
]


cfg.CONF.register_opts(agent_opts, "AGENT")
a_config.register_agent_state_opts_helper(cfg.CONF)
a_config.register_root_helper(cfg.CONF)

ACONF = cfg.CONF.AGENT


class UpdateVLANRequest(object):
    """Used for the async update of the PVIDs on ports."""

    def __init__(self, p_req):
        """Creates a request to update the VLAN.

        :param p_req: The backing ProvisionRequest which provides details
                      about which client LPAR and Network Adapter to update
                      the VLAN on.
        """
        self.p_req = p_req
        self.attempt_count = 0


class PVIDLooper(object):
    """This class is used to monitor and apply update PVIDs to CNAs.

    When Neutron receives a Port Create request, the client CNA needs to have
    the appropriate VLAN applied to it.  However, the port create is usually
    done before the CNA actually exists.

    This class will listen for a period of time, and when the CNA becomes
    available, will update the CNA with the appropriate PVID.
    """

    def __init__(self, agent):
        """Initializes the looper.

        :param agent: The agent running the PVIDLooper
        """
        self.requests = []
        self.agent = agent
        self.adapter = agent.adapter
        self.host_uuid = agent.host_uuid

    def update(self):
        """Performs a loop and updates all of the queued requests."""
        current_requests = copy.copy(self.requests)

        # No requests, do nothing.
        if len(current_requests) == 0:
            return

        # Get the lpar UUIDs up front.
        lpar_uuids = utils.list_lpar_uuids(self.adapter, self.host_uuid)

        # Loop through the current requests.  Try to update the PVIDs, but
        # if we are unable, then increment the attempt count.
        for request in current_requests:
            self._update_req(request, lpar_uuids)

    def _update_req(self, request, lpar_uuids):
        """Attempts to provision a given UpdateVLANRequest.

        :param request: The UpdateVLANRequest.
        :return: True if the request was successfully processed.  False if it
                 was not able to process.
        """
        # Pull the ProvisionRequest off the VLAN Update call.
        p_req = request.p_req
        client_adpts = []

        try:
            if p_req.lpar_uuid in lpar_uuids:
                # Get the adapters just for the VM that the request is for.
                client_adpts = utils.list_cnas(self.adapter, self.host_uuid,
                                               lpar_uuid=p_req.lpar_uuid)
                cna = utils.find_cna_for_mac(p_req.mac_address, client_adpts)
                if cna:
                    # If the PVID does not match, update the CNA.
                    if cna.pvid != p_req.segmentation_id:
                        utils.update_cna_pvid(cna, p_req.segmentation_id)
                    LOG.info(_LI("Sending update device for %s"),
                             p_req.mac_address)
                    self.agent.update_device_up(p_req.rpc_device)
                    self._remove_request(request)
                    return

        except Exception as e:
            LOG.warning(_LW("An error occurred while attempting to update the "
                            "PVID of the virtual NIC."))
            LOG.exception(e)

        # Increment the request count.
        request.attempt_count += 1
        if request.attempt_count >= ACONF.pvid_update_loops:
            # If it had been on the system...this is an error.
            if p_req.lpar_uuid in lpar_uuids:
                self._mark_failed(p_req, client_adpts)

            # Remove the request from the overall queue
            self._remove_request(request)

    def _mark_failed(self, p_req, client_adpts):
        """Marks a provision request as failed."""
        LOG.error(_LE("Unable to update PVID to %(pvid)s for MAC Address "
                      "%(mac)s as there was no valid network adapter found."),
                  {'pvid': p_req.segmentation_id, 'mac': p_req.mac_address})

        # Log additionally the adapters (if any) that were found for the
        # client LPAR.
        count = 0
        for cna in client_adpts:
            LOG.error(_LE("Existing Adapter %(num)d: mac %(mac)s, pvid "
                          "%(pvid)d"), {'num': count, 'mac': cna.mac,
                                        'pvid': cna.pvid})
            count += 1

        # Mark the device down.
        self.agent.update_device_down(p_req.rpc_device)

    def looping_call(self):
        """Runs the update method, but wraps a try/except block around it."""
        while True:
            try:
                self.update()
            except Exception as e:
                # Only log the exception, do not block the processing.
                LOG.exception(e)

            # Sleep for a second.
            time.sleep(1)

    @lockutils.synchronized('pvid_looper_req')
    def _remove_request(self, request):
        self.requests.remove(request)

    @lockutils.synchronized('pvid_looper_req')
    def add(self, request):
        """Adds a new request to the looper utility.

        :param request: A UpdateVLANRequest.
        """
        # The request may already be on the system.  This can happen due to the
        # CNAEventHandler...it may have been invoked a few times from LPAR
        # invalidates...thus making it appear like the request came through
        # a few times.
        #
        # We should look at the existing queue, and only add to it if we do
        # not have one with a similar to it.
        contains = False
        for existing_req in self.requests:
            if existing_req.p_req == request.p_req:
                contains = True
                break

        if not contains:
            self.requests.append(request)

    @property
    @lockutils.synchronized('pvid_looper_req')
    def pending_vlans(self):
        """Returns the set of pending VLAN updates.

        :return: Set of unique VLAN ids from within the pending requests.
        """
        return {x.p_req.segmentation_id for x in self.requests}


class SharedEthernetNeutronAgent(agent_base.BasePVMNeutronAgent):
    """
    Provides VLAN networks for the PowerVM platform that run accross the
    Shared Ethernet within the Virtual I/O Servers.  Designed to be compatible
    with the ML2 Neutron Plugin.
    """

    def __init__(self):
        """Constructs the agent."""
        name = 'networking-powervm-sharedethernet-agent'
        agent_type = p_const.AGENT_TYPE_PVM_SEA

        super(SharedEthernetNeutronAgent, self).__init__(name, agent_type)

        # A looping utility that updates asynchronously the PVIDs on the
        # Client Network Adapters (CNAs)
        self.pvid_updater = PVIDLooper(self)
        eventlet.spawn_n(self.pvid_updater.looping_call)

    def parse_bridge_mappings(self):
        return utils.parse_sea_mappings(self.adapter, self.host_uuid,
                                        ACONF.bridge_mappings)

    def build_prov_requests_from_server(self):
        """Builds provisioning requests from the server.

        The server may detect that a new port has been built on its system.
        This method provides the agent implementations to detect this, and
        return a ProvisionRequest object that will be passed into the
        attempt_provision method.

        This method is not required to be implemented by agent implementations.
        """
        return self._cna_event_handler.get_queue()

    def heal_and_optimize(self, is_boot, prov_reqs, lpar_uuids, overall_cnas):
        """Heals the system's network bridges and optimizes.

        Will query neutron for all the ports in use on this host.  Ensures that
        all of the VLANs needed for those ports are available on the correct
        network bridge.

        Finally, it optimizes the system by removing any VLANs that may no
        longer be required.  The VLANs that are removed must meet the following
        conditions:
         - Are not in use by ANY virtual machines on the system.  OpenStack
           managed or not.
         - Are not part of the primary load group on the Network Bridge.

        :param is_boot: Indicates if this is the first call on boot up of the
                        agent.
        :param prov_reqs: A list of ProvisionRequest objects that represent
                          the Neutron ports that should exist on this system.
                          It may include ports that have already been
                          provisioned.  This method should make sure it calls
                          update_device_up/down afterwards.
        :param lpar_uuids: A list of the VM UUIDs for the REST API.
        :param overall_cnas: A list of the systems Client Network Adapters.
        """
        # Dictionary of the required VLANs on the Network Bridge
        nb_req_vlans = {}
        nb_wraps = utils.list_bridges(self.adapter, self.host_uuid)
        for nb_wrap in nb_wraps:
            nb_req_vlans[nb_wrap.uuid] = set()

        # Call down to the provision.  This will call device up on the
        # requests.
        self.provision_devices(prov_reqs)

        # Make sure that the provision requests VLAN is captured in the
        # nb_req_vlans list...so that the VLAN is not accidentally removed.
        for req in prov_reqs:
            nb_uuid, req_vlan = self._get_nb_and_vlan(req.rpc_device,
                                                      emit_warnings=False)
            nb_req_vlans[nb_uuid].add(req_vlan)

        # We should clean up old VLANs as well.  However, we only want to clean
        # up old VLANs that are not in use by ANYTHING in the system.
        #
        # The first step is to identify the VLANs that are needed.  That can
        # be done by extending our nb_req_vlans map.
        #
        # We first extend that map by listing all the VMs on the system
        # (whether managed by OpenStack or not) and then seeing what Network
        # Bridge uses them.
        vswitch_map = utils.get_vswitch_map(self.adapter, self.host_uuid)
        for client_adpt in overall_cnas:
            nb = utils.find_nb_for_cna(nb_wraps, client_adpt, vswitch_map)
            # Could occur if a system is internal only.
            if nb is None:
                LOG.debug("Client Adapter with mac %s is internal only.",
                          client_adpt.mac)
                continue

            # Make sure that it is on the nb_req_vlans list, as it is now
            # considered required.
            nb_req_vlans[nb.uuid].add(client_adpt.pvid)

            # Extend for each additional vlans as well
            for addl_vlan in client_adpt.tagged_vlans:
                nb_req_vlans[nb.uuid].add(addl_vlan)

        # We will have a list of CNAs that are not yet created, but are pending
        # provisioning from Nova.  Keep track of those so that we don't tear
        # those off the SEA.
        pending_vlans = self.pvid_updater.pending_vlans

        # The list of required VLANs on each network bridge also includes
        # everything on the primary VEA.
        for nb in nb_wraps:
            prim_ld_grp = nb.load_grps[0]
            vlans = [prim_ld_grp.pvid]
            vlans.extend(prim_ld_grp.tagged_vlans)
            for vlan in vlans:
                nb_req_vlans[nb.uuid].add(vlan)

        # If the configuration is set.
        if ACONF.automated_powervm_vlan_cleanup:
            # Loop through and remove VLANs that are no longer needed.
            for nb in nb_wraps:
                # Join the required vlans on the network bridge (already in
                # use) with the pending VLANs.
                req_vlans = nb_req_vlans[nb.uuid] | pending_vlans

                # Get ALL the VLANs on the bridge
                existing_vlans = set(nb.list_vlans())

                # To determine the ones no longer needed, subtract from all the
                # VLANs the ones that are no longer needed.
                vlans_to_del = existing_vlans - req_vlans
                for vlan_to_del in vlans_to_del:
                    LOG.warning(_LW("Cleaning up VLAN %(vlan)s from the "
                                    "system. It is no longer in use."),
                                {'vlan': vlan_to_del})
                    net_br.remove_vlan_from_nb(self.adapter, self.host_uuid,
                                               nb.uuid, vlan_to_del)

    def provision_devices(self, requests):
        """Will ensure that the VLANs are on the NBs for the edge devices.

        Takes in a set of ProvisionRequests.  From those devices, determines
        the correct network bridges and their appropriate VLANs.  Then calls
        down to the pypowervm API to ensure that the required VLANs are
        on the appropriate ports.

        Will also ensure that the client side adapter is updated with the
        correct VLAN.

        :param requests: A list of ProvisionRequest objects.
        """
        nb_to_vlan = {}
        for p_req in requests:
            # Break the ports into their respective lists broken down by
            # Network Bridge.
            nb_uuid, vlan = self._get_nb_and_vlan(p_req.rpc_device,
                                                  emit_warnings=True)

            # A warning message will be printed to user if this were to occur
            if nb_uuid is None:
                continue

            if nb_to_vlan.get(nb_uuid) is None:
                nb_to_vlan[nb_uuid] = set()

            nb_to_vlan[nb_uuid].add(vlan)

        # For each bridge, make sure the VLANs are serviced.
        for nb_uuid in nb_to_vlan.keys():
            net_br.ensure_vlans_on_nb(self.adapter, self.host_uuid, nb_uuid,
                                      nb_to_vlan.get(nb_uuid))

        # Now that the bridging is complete, loop through the devices again
        # and kick off the PVID update on the client devices.  This should
        # not be done until the vlan is on the network bridge.  Otherwise the
        # port state in the backing neutron server could be out of sync.
        for p_req in requests:
            self.pvid_updater.add(UpdateVLANRequest(p_req))
        LOG.debug('Successfully provisioned new devices.')

    def _get_nb_and_vlan(self, dev, emit_warnings=False):
        """Parses bridge mappings to find a match for the device passed in.
        :param dev: Neutron device to find a match for
        :param emit_warnings: (Optional) Defaults to False.  If true, will emit
                              a warning if the configuration is off.
        :return: UUID of the NetBridge
        :return: vlan for the neutron device
        """
        nb_uuid = self.br_map.get(dev.get('physical_network'))
        if not nb_uuid and emit_warnings:
            LOG.warning(_LW("Unable to determine the Network Bridge (Shared "
                            "Ethernet Adapter) for physical network %s.  Will "
                            "be unable to determine appropriate provisioning "
                            "action."), dev.get('physical_network'))
        return nb_uuid, dev.get('segmentation_id')


def main():
    # Read in the command line args
    n_config.init(sys.argv[1:])
    n_config.setup_logging()

    # Build then run the agent
    agent = SharedEthernetNeutronAgent()
    LOG.info(_LI("Shared Ethernet Agent initialized and running"))
    agent.rpc_loop()


if __name__ == "__main__":
    main()



# Copyright 2014 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_powervm.plugins.ml2.drivers import mech_pvm_sea as m_pvm
from networking_powervm.tests.unit.plugins.ibm.powervm import base

from neutron.plugins.ml2 import driver_api as api


class TestPvmMechDriver(base.BasePVMTestCase):

    def setUp(self):
        super(TestPvmMechDriver, self).setUp()
        self.mech_drv = m_pvm.PvmSEAMechanismDriver()

    @mock.patch('networking_powervm.plugins.ml2.drivers.mech_pvm_sea.'
                'PvmSEAMechanismDriver.get_mappings')
    def test_check_segment_for_agent(self, mappings):
        """Validates that the VLAN type is supported by the agent."""
        # Only test the flow where we have custom code.
        mappings.return_value = {}

        fake_segment = {api.NETWORK_TYPE: 'vlan'}
        self.assertTrue(self.mech_drv.check_segment_for_agent(fake_segment,
                                                              None))

        bad_segment = {api.NETWORK_TYPE: 'gre'}
        self.assertFalse(self.mech_drv.check_segment_for_agent(bad_segment,
                                                               None))

    @mock.patch('neutron.plugins.ml2.drivers.mech_agent.'
                'SimpleAgentMechanismDriverBase.'
                'try_to_bind_segment_for_agent', return_value=True)
    def test_try_to_bind_segment_for_agent(self, try_bind):
        fake_segment = {api.NETWORK_TYPE: 'vlan', api.SEGMENTATION_ID: '1000',
                        api.PHYSICAL_NETWORK: 'default'}
        fake_context = mock.MagicMock()
        self.mech_drv.rpc_publisher = mock.MagicMock()
        self.mech_drv.try_to_bind_segment_for_agent(
            fake_context, fake_segment, None)
        self.mech_drv.rpc_publisher.port_update.assert_called_with(
            fake_context._plugin_context, fake_context._port,
            'vlan', '1000', 'default')



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import eventlet
eventlet.monkey_patch()

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall

from neutron.agent.common import config as a_config
from neutron.agent import rpc as agent_rpc
from neutron.common import constants as q_const
from neutron.common import topics
from neutron import context as ctx
from pypowervm import adapter as pvm_adpt
from pypowervm.helpers import log_helper as log_hlp
from pypowervm.helpers import vios_busy as vio_hlp
from pypowervm import util as pvm_util
from pypowervm.utils import uuid as pvm_uuid

from networking_powervm._i18n import _
from networking_powervm._i18n import _LE
from networking_powervm._i18n import _LI
from networking_powervm._i18n import _LW
from networking_powervm.plugins.ibm.agent.powervm import utils

import time


LOG = logging.getLogger(__name__)


agent_opts = [
    cfg.IntOpt('exception_interval', default=5,
               help=_("The number of seconds agent will wait between "
                      "polling when exception is caught")),
    cfg.IntOpt('polling_interval', default=2,
               help=_("The number of seconds the agent will wait between "
                      "polling for local device changes.")),
    cfg.IntOpt('heal_and_optimize_interval', default=1800,
               help=_('The number of seconds the agent should wait between '
                      'heal/optimize intervals.  Should be higher than the '
                      'polling_interval as it runs in the nearest polling '
                      'loop.'))
]

cfg.CONF.register_opts(agent_opts, "AGENT")
a_config.register_agent_state_opts_helper(cfg.CONF)
a_config.register_root_helper(cfg.CONF)

ACONF = cfg.CONF.AGENT


class PVMPluginApi(agent_rpc.PluginApi):
    pass


class PVMRpcCallbacks(object):
    """
    Provides call backs (as defined in the setup_rpc method within the
    appropriate Neutron Agent class) that will be invoked upon certain
    actions from the controller.
    """

    # This agent supports RPC Version 1.0.  Though agents don't boot unless
    # 1.1 or higher is specified now.
    # For reference:
    #  1.0 Initial version
    #  1.1 Support Security Group RPC
    #  1.2 Support DVR (Distributed Virtual Router) RPC
    RPC_API_VERSION = '1.1'

    def __init__(self, agent):
        """
        Creates the call back.  Most of the call back methods will be
        delegated to the agent.

        :param agent: The owning agent to delegate the callbacks to.
        """
        super(PVMRpcCallbacks, self).__init__()
        self.agent = agent

    def port_update(self, context, **kwargs):
        port = kwargs['port']
        self.agent._update_port(port)
        LOG.debug("port_update RPC received for port: %s", port['id'])

    def network_delete(self, context, **kwargs):
        network_id = kwargs.get('network_id')
        LOG.debug("network_delete RPC received for network: %s", network_id)


class ProvisionRequest(object):
    """A request for a Neutron Port to be provisioned.

    The RPC device details provide some additional details that the port does
    not necessarily have, and vice versa.  This meshes together the required
    aspects into a single element.
    """

    def __init__(self, device_detail, lpar_uuid, client_type='vea'):
        self.segmentation_id = device_detail.get('segmentation_id')
        self.physical_network = device_detail.get('physical_network')
        self.mac_address = device_detail.get('mac_address')
        self.device_owner = device_detail.get('device_owner')
        self.rpc_device = device_detail
        self.lpar_uuid = lpar_uuid
        self.client_type = client_type

    def __eq__(self, other):
        if not isinstance(other, ProvisionRequest):
            return False

        # Really just need to check the lpar_uuid and mac.  The rest should
        # be static and identical.
        return (other.mac_address == self.mac_address and
                other.lpar_uuid == self.lpar_uuid)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        # Mac addresses should not collide.  This should be sufficient for a
        # hash.  The equals will go just a bit further.
        return hash(self.mac_address)


class CNAEventHandler(pvm_adpt.EventHandler):
    """Listens for Events from the PowerVM API that could be network events.

    This event handler will be invoked by the PowerVM API when something occurs
    on the system.  This event handler will determine if it could have been
    related to a network change.  If so, then it will add a ProvisionRequest
    to the processing queue.
    """

    def __init__(self, agent):
        self.agent = agent
        self.adapter = self.agent.adapter
        self.host_uuid = self.agent.host_uuid
        self.prov_req_queue = []

    @lockutils.synchronized('cna_request_queue')
    def process(self, events):
        for uri, action in events.items():
            if action in ['add', 'invalidate']:
                self.prov_req_queue.extend(self._prov_reqs_for_uri(uri))

    def _prov_reqs_for_uri(self, uri):
        """Returns set of ProvisionRequests for a URI.

        When the API indicates that a URI is invalid, it will return a
        List of ProvisionRequests for a given URI.  If the URI is not valid
        for a ClientNetworkAdapter (CNA) then an empty list will be returned.
        """
        try:
            if not pvm_util.is_instance_path(uri):
                return []
        except Exception:
            LOG.warning(_LW('Unable to parse URI %s for provision request '
                            'assessment.'), uri)
            return []

        # The event queue will only return URI's for 'root like' objects.
        # This is essentially just the LogicalPartition, you can't get the
        # ClientNetworkAdapter.  So if we find an add/invalidate for the
        # LogicalPartition, we'll get all the CNAs.
        #
        # This check will throw out everything that doesn't include the
        # LogicalPartition's
        uuid = pvm_util.get_req_path_uuid(uri, preserve_case=True)
        if not uri.endswith('LogicalPartition/' + uuid):
            return []

        # For the LPAR, get the CNAs.
        cna_wraps = utils.list_cnas(self.adapter, self.host_uuid, uuid)
        resp = []
        for cna_w in cna_wraps:
            # Build a provision request for each type
            device_mac = utils.norm_mac(cna_w.mac)
            device_detail = self.agent.get_device_details(device_mac)

            # A device detail will always come back...even if neutron has
            # no idea what the port is.  This WILL happen for PowerVM, maybe
            # an event for the mgmt partition or the secure RMC VIF.  We can
            # detect if Neutron has full device details by simply querying for
            # the mac from the device_detail
            if not device_detail.get('mac_address'):
                continue

            # Must be good!
            resp.append(ProvisionRequest(device_detail, uuid))
        return resp

    @lockutils.synchronized('cna_request_queue')
    def get_queue(self):
        resp = copy.copy(self.prov_req_queue)
        self.prov_req_queue = []
        return resp


class BasePVMNeutronAgent(object):
    """Baseline PowerVM Neutron Agent class for extension.

    The ML2 agents have a common RPC polling framework and API callback
    mechanism.  This class provides the baseline so that other children
    classes can extend and focus on their specific functions rather than
    integration with the RPC server.
    """

    def __init__(self, binary_name, agent_type):
        # Create the utility class that enables work against the Hypervisors
        # Shared Ethernet NetworkBridge.
        self.setup_adapter()

        # Get the bridge mappings
        self.br_map = self.parse_bridge_mappings()

        self.agent_state = {
            'binary': binary_name,
            'host': cfg.CONF.host,
            'topic': q_const.L2_AGENT_TOPIC,
            'configurations': {'bridge_mappings': self.br_map},
            'agent_type': agent_type,
            'start_flag': True}
        # A list of ports that maintains the list of current 'modified' ports
        self.updated_ports = []

        # Set Up RPC to Server
        self.setup_rpc()

    def setup_adapter(self):
        """Configures the pypowervm adapter and utilities."""
        # Build the adapter.  May need to attempt the connection multiple times
        # in case the REST server is starting.
        session = pvm_adpt.Session(conn_tries=300)
        self.adapter = pvm_adpt.Adapter(
            session, helpers=[log_hlp.log_helper,
                              vio_hlp.vios_busy_retry_helper])
        self.host_uuid = utils.get_host_uuid(self.adapter)

        # Add an event handler to the session.
        evt_listener = session.get_event_listener()
        self._cna_event_handler = CNAEventHandler(self)
        evt_listener.subscribe(self._cna_event_handler)

    def parse_bridge_mappings(self):
        """This method should return the bridge mappings dictionary.

        The pypowervm adapter will be initialized before this method is called.
        """
        raise NotImplementedError()

    def setup_rpc(self):
        """Registers the RPC consumers for the plugin."""
        self.agent_id = 'sea-agent-%s' % cfg.CONF.host
        self.topic = topics.AGENT
        self.plugin_rpc = PVMPluginApi(topics.PLUGIN)
        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.PLUGIN)

        self.context = ctx.get_admin_context_without_session()

        # Defines what will be listening for incoming events from the
        # controller.
        self.endpoints = [PVMRpcCallbacks(self)]

        # Define the listening consumers for the agent.  ML2 only supports
        # these two update types.
        consumers = [[topics.PORT, topics.UPDATE],
                     [topics.NETWORK, topics.DELETE]]

        self.connection = agent_rpc.create_consumers(self.endpoints,
                                                     self.topic,
                                                     consumers)

        # Report interval is for the agent health check.
        report_interval = cfg.CONF.AGENT.report_interval
        if report_interval:
            hb = loopingcall.FixedIntervalLoopingCall(self._report_state)
            hb.start(interval=report_interval)

    def _report_state(self):
        """
        Reports the state of the agent back to the controller.  Controller
        knows that if a response isn't provided in a certain period of time
        then the agent is dead.  This call simply tells the controller that
        the agent is alive.
        """
        # TODO(thorst) provide some level of devices connected to this agent.
        try:
            device_count = 0
            self.agent_state.get('configurations')['devices'] = device_count
            self.state_rpc.report_state(self.context,
                                        self.agent_state)
            self.agent_state.pop('start_flag', None)
        except Exception:
            LOG.exception(_LE("Failed reporting state!"))

    def update_device_up(self, device):
        """Calls back to neutron that a device is alive."""
        self.plugin_rpc.update_device_up(self.context, device['device'],
                                         self.agent_id, cfg.CONF.host)

    def update_device_down(self, device):
        """Calls back to neutron that a device is down."""
        self.plugin_rpc.update_device_down(self.context, device['device'],
                                           self.agent_id, cfg.CONF.host)

    def get_device_details(self, device_mac):
        """Returns a neutron device for a given mac address.

        :param device_mac: The neutron mac addresses for the device to get.
        :return: The device from neutron.
        """
        return self.plugin_rpc.get_device_details(self.context, device_mac,
                                                  self.agent_id)

    def get_devices_details_list(self, device_macs):
        """Returns list of neutron devices for a list of mac addresses.

        :param device_macs: List of neutron mac addresses for the devices to
                            get.
        :return: The list of devices from neutron.
        """
        return self.plugin_rpc.get_devices_details_list(
            self.context, device_macs, self.agent_id)

    def _update_port(self, port):
        """Invoked to indicate that a port has been updated within Neutron."""
        LOG.info(_LI('Neutron API indicated port update for %(mac)s.  '
                     'Checking if hosted by this system.'),
                 {'mac': port.get('mac_address')})
        self.updated_ports.append(port)

    def _list_updated_ports(self):
        """
        Will return (and then reset) the list of updated ports received
        from the system.
        """
        ports = copy.copy(self.updated_ports)
        self.updated_ports = []
        return ports

    def _build_system_prov_requests(self):
        """Builds ProvisionRequests for the entire system.

        It is expected this is used in the heal_and_optimize code path.  After
        these devices are pulled, something will need to reset their states
        back to 'up'.  That is the requirement of the implementing class.
        """
        prov_reqs = []

        # Get all the LPARs...we need to match up the client adapters with a
        # given device.
        lpar_uuids = utils.list_lpar_uuids(self.adapter, self.host_uuid)
        overall_cnas = []
        lpar_cna_map = {}
        for lpar_uuid in lpar_uuids:
            lpar_cna_map[lpar_uuid] = utils.list_cnas(
                self.adapter, self.host_uuid, lpar_uuid=lpar_uuid)
            overall_cnas.extend(lpar_cna_map[lpar_uuid])

        # Get all the devices that Neutron knows for this host.  Note that
        # we pass in all of the macs on the system.  For VMs that neutron does
        # not know about, we get back an empty structure with just the mac.
        client_macs = [utils.norm_mac(x.mac) for x in overall_cnas]
        devs = self.get_devices_details_list(client_macs)

        # Build out all of the devices that we have available to us.  Some
        # may be on the system but not part of OpenStack.  Those get ignored.
        for lpar_uuid in lpar_cna_map.keys():
            for cna in lpar_cna_map[lpar_uuid]:
                dev = self._find_dev(cna, devs)
                if dev:
                    prov_reqs.append(ProvisionRequest(dev, lpar_uuid))

        return prov_reqs, lpar_uuids, overall_cnas

    def _find_dev(self, cna, devs):
        cna_mac = utils.norm_mac(cna.mac)
        for dev in devs:
            # A device detail will always come back...even if neutron has
            # no idea what the port is.  This WILL happen for PowerVM, maybe
            # an event for the mgmt partition or the secure RMC VIF.  We can
            # detect if Neutron has full device details by simply querying for
            # the mac from the device_detail
            if not dev.get('mac_address'):
                continue

            if utils.norm_mac(dev.get('mac_address')) == cna_mac:
                return dev
        return None

    def heal_and_optimize(self, is_boot, prov_reqs, lpar_uuids, overall_cnas):
        """Ensures that the bridging supports all the needed ports.

        This method is invoked periodically (not on every RPC loop).  Its
        purpose is to ensure that the bridging supports every client VM
        properly.  If possible, it should also optimize the connections.

        :param is_boot: Indicates if this is the first call on boot up of the
                        agent.
        :param prov_reqs: A list of ProvisionRequest objects that represent
                          the Neutron ports that should exist on this system.
                          It may include ports that have already been
                          provisioned.  This method should make sure it calls
                          update_device_up/down afterwards.
        :param lpar_uuids: A list of the VM UUIDs for the REST API.
        :param overall_cnas: A list of the systems Client Network Adapters.
        """
        raise NotImplementedError()

    def provision_devices(self, requests):
        """Invoked when a set of new Neutron ports has been detected.

        This method should provision the bridging for the new devices

        Must be implemented by a subclass.

        The subclass implementation may be non-blocking.  This means, if it
        will take a very long time to provision, or has a dependency on
        another action (ex. client VIF needs to be created), then it should
        run in a separate worker thread.

        Because of the non-blocking nature of the method, it is required that
        the child class updates the device state upon completion of the device
        provisioning.  This can be done with the agent's
        update_device_up/_down methods.

        :param requests: A list of ProvisionRequest objects.
        """
        raise NotImplementedError()

    def rpc_loop(self):
        """
        Runs a check periodically to determine if new ports were added or
        removed.  Will call down to appropriate methods to determine correct
        course of action.
        """

        loop_timer = float(0)
        loop_interval = float(ACONF.heal_and_optimize_interval)
        first_loop = True

        while True:
            try:
                # If the loop interval has passed, heal and optimize
                if time.time() - loop_timer > loop_interval:
                    LOG.debug("Performing heal and optimization of system.")
                    prov_reqs, lpar_uuids, overall_cnas =\
                        self._build_system_prov_requests()
                    self.heal_and_optimize(first_loop, prov_reqs, lpar_uuids,
                                           overall_cnas)
                    first_loop = False
                    loop_timer = time.time()

                # Determine if there are new ports requested from neutron
                n_prov_reqs = self.build_prov_requests_from_neutron()

                # Get provision requests from the server
                s_prov_reqs = self.build_prov_requests_from_server()

                # Get all of the provision requests, but remove any duplicates.
                # A duplicate could occur if the server and neutron both threw
                # the same port request.
                tot_prov_reqs = n_prov_reqs + s_prov_reqs
                tot_prov_reqs = list(set(tot_prov_reqs))

                # If there are no updated ports, just sleep and re-loop
                if not tot_prov_reqs:
                    LOG.debug("No changes, sleeping %d seconds.",
                              ACONF.polling_interval)
                    time.sleep(ACONF.polling_interval)
                    continue

                # Provision the ports on the Network Bridge.
                self.attempt_provision(tot_prov_reqs)

            except Exception as e:
                LOG.exception(e)
                LOG.warning(_LW("Error has been encountered and logged.  The "
                                "agent will retry again."))
                # sleep for a while and re-loop
                time.sleep(ACONF.exception_interval)

    def build_prov_requests_from_neutron(self):
        """Builds the provisioning requests from the Neutron Server.

        The Neutron Server may have updated ports.  These port requests will
        be sent down to the agent as a ProvisionRequest.

        :return: A list of the ProvisionRequests that have come from Neutron.
        """
        # Convert the ports to devices.
        u_ports = self._list_updated_ports()
        dev_list = [x.get('mac_address') for x in u_ports]
        devices = self.get_devices_details_list(dev_list)

        # Build the network devices
        resp = []
        for port in u_ports:
            port_uuid = port.get('id')

            # Make sure we have a UUID
            if port_uuid is None:
                continue

            # Make sure the binding host matches this agent.  Otherwise it is
            # meant to provision on another agent.
            if port.get('binding:host_id') != cfg.CONF.host:
                continue

            for dev in devices:
                # If the device's id (really the port uuid) doesn't match,
                # ignore it.
                dev_pid = dev.get('port_id')
                if dev_pid is None or port_uuid != dev_pid:
                    continue

                # Valid request.  Add it
                device_id = port.get('device_id')
                lpar_uuid = pvm_uuid.convert_uuid_to_pvm(device_id).upper()
                resp.append(ProvisionRequest(dev, lpar_uuid))
        return resp

    def build_prov_requests_from_server(self):
        """Builds provisioning requests from the server.

        The server may detect that a new port has been built on its system.
        This method provides the agent implementations to detect this, and
        return a ProvisionRequest object that will be passed into the
        attempt_provision method.

        This method is not required to be implemented by agent implementations.
        """
        return self._cna_event_handler.get_queue()

    def attempt_provision(self, provision_reqs):
        """Attempts the provisioning of ports.

        This method will attempt to provision a set of ports (by wrapping
        around the provision_ports method).  If there are issues with
        provisioning the ports, this method will update the status in the
        backing Neutron server.

        :param provision_reqs: The list of ports to provision.
        """
        try:
            LOG.debug("Provisioning ports for mac addresses [ %s ]" %
                      ' '.join([x.mac_address for x in provision_reqs]))
            self.provision_devices(provision_reqs)
        except Exception:
            # Set the state of the device as 'down'
            for p_req in provision_reqs:
                self.update_device_down(p_req.rpc_device)

            # Reraise the exception
            raise



# Copyright 2014 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests import base

import os
import shutil


class BasePVMTestCase(base.BaseTestCase):
    """The base PowerVM Test case."""

    def setUp(self):
        super(BasePVMTestCase, self).setUp()

        # We need to try to copy over the policy.json.  Some neutron
        # modules load it, but since we start in a different location
        # we lose its context...  Copying will bring it in.
        policy_json, to = self._get_policy_paths()
        if not os.path.exists(to):
            shutil.copyfile(policy_json, to)

    def tearDown(self):
        super(BasePVMTestCase, self).tearDown()

        # Remove the policy now that it is no longer used.
        policy_json, to = self._get_policy_paths()
        if os.path.exists(to):
            os.remove(to)

    def _get_policy_paths(self):
        """
        Returns the source policy path from neutron and a target path
        to store the file in temporarily for the tests.
        """
        # Start with the source path.
        tests = os.path.split(base.__file__)[0]
        neutron_src = os.path.split(tests)[0]
        neutron = os.path.split(neutron_src)[0]
        policy_json = os.path.join(neutron, 'etc/policy.json')

        # Get the copy to path
        home_path = os.path.abspath(os.path.expanduser('~'))
        to = os.path.join(home_path, 'policy.json')

        # return the two
        return policy_json, to



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.common import exceptions

from networking_powervm._i18n import _LE


class MultipleHostsFound(exceptions.NeutronException):
    message = _LE("Expected exactly one host; found %(host_count)d.")


class NoNetworkBridges(exceptions.NeutronException):
    message = _LE('There are no network bridges (Shared Ethernet Adapters) on '
                  'the system.  Can not start the Neutron agent.')


class MultiBridgeNoMapping(exceptions.NeutronException):
    message = _LE('The system has more than one network bridge, but the '
                  'bridge_mappings have not been specified.  Please configure '
                  'the bridge_mappings before proceeding.')


class DeviceNotFound(exceptions.NeutronException):
    message = _LE('Device %(dev)s on Virtual I/O Server %(vios)s was not '
                  'found.  Unable to set up physical network %(phys_net)s.')



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from oslo_config import cfg
from pypowervm.helpers import log_helper as log_hlp
from pypowervm.helpers import vios_busy as vio_hlp
from pypowervm.tests import test_fixtures as pvm_fx

from networking_powervm.plugins.ibm.agent.powervm import agent_base
from networking_powervm.tests.unit.plugins.ibm.powervm import base


class FakeExc(Exception):
    pass


def FakeNPort(mac, segment_id, phys_network):
    return {'mac_address': mac, 'segmentation_id': segment_id,
            'physical_network': phys_network}


class TestAgentBaseInit(base.BasePVMTestCase):
    """A test class to validate the set up of the agent with the API.

    This is typically mocked out in fixtures otherwise.
    """

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'utils.get_host_uuid')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'CNAEventHandler')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'BasePVMNeutronAgent.setup_rpc')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'BasePVMNeutronAgent.parse_bridge_mappings')
    @mock.patch('pypowervm.adapter.Adapter')
    @mock.patch('pypowervm.adapter.Session')
    def test_setup_adapter(self, mock_session, mock_adapter,
                           mock_parse_mappings, mock_setup_rpc,
                           mock_evt_handler, mock_host_uuid):
        # Set up the mocks.
        mock_evt_listener = (mock_session.return_value.get_event_listener.
                             return_value)
        mock_evt_handler.return_value = 'evt_hdlr'
        mock_host_uuid.return_value = 'host_uuid'

        # Setup and invoke
        neut_agt = agent_base.BasePVMNeutronAgent('bin', 'type')

        # Validate
        mock_session.assert_called_once_with(conn_tries=300)
        mock_adapter.assert_called_once_with(
            mock_session.return_value,
            helpers=[log_hlp.log_helper, vio_hlp.vios_busy_retry_helper])
        self.assertEqual('host_uuid', neut_agt.host_uuid)
        mock_evt_listener.subscribe.assert_called_once_with('evt_hdlr')


class TestAgentBase(base.BasePVMTestCase):

    def build_test_agent(self):
        """Builds a simple test agent."""
        self.adpt = self.useFixture(
            pvm_fx.AdapterFx(traits=pvm_fx.LocalPVMTraits)).adpt

        with mock.patch('networking_powervm.plugins.ibm.agent.powervm.'
                        'agent_base.BasePVMNeutronAgent.setup_adapter'),\
                mock.patch('networking_powervm.plugins.ibm.agent.powervm.'
                           'agent_base.BasePVMNeutronAgent.setup_rpc'),\
                mock.patch('networking_powervm.plugins.ibm.agent.powervm.'
                           'agent_base.BasePVMNeutronAgent.'
                           'parse_bridge_mappings'):
            agent = agent_base.BasePVMNeutronAgent('binary_name', 'agent_type')
            agent.context = mock.Mock()
            agent.host_uuid = 'host_uuid'
            agent.agent_id = 'pvm'
            agent.plugin_rpc = mock.MagicMock()
            agent.adapter = self.adpt
        return agent

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'BasePVMNeutronAgent.get_devices_details_list')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_cnas')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_lpar_uuids')
    def test_build_system_prov_requests(self, mock_lpar_list, mock_list_cnas,
                                        mock_dev_details):
        agent = self.build_test_agent()

        # Mock data
        mock_lpar_list.return_value = ['uuid1', 'uuid2', 'uuid3']
        mock_list_cnas.side_effect = [[mock.Mock(mac='aabbccddeefd')],
                                      [mock.Mock(mac='aabbccddeefe')],
                                      [mock.Mock(mac='aabbccddeeff')]]
        mock_dev_details.return_value = [
            {'device': 'aa:bb:cc:dd:ee:fd'}, {'mac_address': 'aabbccddeefe'},
            {'mac_address': 'aa:bb:cc:dd:ee:ff'}]

        # Run the method
        prov_reqs, lpar_uuids, cnas = agent._build_system_prov_requests()

        # Validation
        self.assertEqual(2, len(prov_reqs))
        self.assertEqual(['uuid1', 'uuid2', 'uuid3'], lpar_uuids)
        self.assertEqual(3, len(cnas))

    def test_find_dev(self):
        agent = self.build_test_agent()

        mock_cna = mock.Mock(mac='aa:bb:cc:dd:ee:ff')

        # This is what a device looks like if you call neutron for a port
        # that exists on the system, but not in Neutron itself.
        mock_dev1 = {'device': 'aa:bb:cc:dd:ee:ff'}

        # This is a subset of the real data you get back for a proper port.
        mock_dev2 = {'mac_address': 'aabbccddeeff'}

        self.assertEqual(mock_dev2,
                         agent._find_dev(mock_cna, [mock_dev1, mock_dev2]))

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'BasePVMNeutronAgent.provision_devices')
    def test_attempt_provision(self, mock_provision):
        """Tests a successful 'attempt_provision' invocation."""
        agent = self.build_test_agent()
        provision_reqs = [mock.Mock(rpc_device='a', mac_address='a'),
                          mock.Mock(rpc_device='b', mac_address='b'),
                          mock.Mock(rpc_device='c', mac_address='c')]

        # Invoke the test method.
        agent.attempt_provision(provision_reqs)

        # Validate the provision was invoked.
        mock_provision.assert_called_with(provision_reqs)

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'BasePVMNeutronAgent.update_device_down')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'BasePVMNeutronAgent.provision_devices')
    def test_attempt_provision_failure(self, mock_provision, mock_dev_down):
        """Tests a failed 'attempt_provision' invocation."""
        agent = self.build_test_agent()

        devs = [mock.Mock(), mock.Mock(), mock.Mock()]
        agent.plugin_rpc.get_devices_details_list.return_value = devs

        # Trigger some failure
        mock_provision.side_effect = FakeExc()
        provision_reqs = [mock.Mock(rpc_device='a', mac_address='a'),
                          mock.Mock(rpc_device='b', mac_address='b'),
                          mock.Mock(rpc_device='c', mac_address='c')]

        # Invoke the test method.
        self.assertRaises(FakeExc, agent.attempt_provision, provision_reqs)

        # Validate the provision was invoked, but failed.
        mock_provision.assert_called_with(provision_reqs)
        self.assertEqual(3, mock_dev_down.call_count)

    @mock.patch('pypowervm.utils.uuid.convert_uuid_to_pvm')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'BasePVMNeutronAgent._list_updated_ports')
    def test_build_prov_requests_from_neutron(self, mock_list_uports,
                                              mock_pvid_convert):
        # Do a base check
        agent = self.build_test_agent()
        mock_list_uports.return_value = []
        self.assertEqual([], agent.build_prov_requests_from_neutron())

        cfg.CONF.set_override('host', 'fake_host')

        def build_port(pid, use_good_host=True):
            if use_good_host:
                return {'id': pid, 'binding:host_id': 'fake_host'}
            else:
                return {'id': pid, 'binding:host_id': 'bad_fake_host'}

        # Only 2 should be created
        mock_list_uports.return_value = [build_port('1'), {}, build_port('2'),
                                         build_port('4', use_good_host=False)]
        devs = [{'port_id': '3'}, {}, {'port_id': '1'}, {'port_id': '2'}]
        agent.plugin_rpc.get_devices_details_list.return_value = devs

        resp = agent.build_prov_requests_from_neutron()
        self.assertEqual(2, len(resp))


class TestProvisionRequest(base.BasePVMTestCase):

    def build_dev(self, segmentation_id, mac):
        return {'segmentation_id': segmentation_id, 'mac_address': mac,
                'physical_network': 'default', 'device_owner': 'nova:compute'}

    def build_preq(self, segmentation_id, mac):
        return agent_base.ProvisionRequest(
            self.build_dev(segmentation_id, mac), '1')

    def test_duplicate_removal(self):
        reqs = [self.build_preq(1, 'a'), self.build_preq(1, 'a'),
                self.build_preq(1, 'b'), self.build_preq(1, 'b'),
                self.build_preq(2, 'c'), self.build_preq(3, 'd')]
        reqs = list(set(reqs))

        # Make sure that the list is properly reduced.
        self.assertEqual(4, len(reqs))
        expected = [self.build_preq(1, 'a'), self.build_preq(1, 'b'),
                    self.build_preq(2, 'c'), self.build_preq(3, 'd')]
        for needle in expected:
            self.assertIn(needle, reqs)


class TestCNAEventHandler(base.BasePVMTestCase):
    """Validates that the CNAEventHandler can be invoked properly."""

    def setUp(self):
        super(TestCNAEventHandler, self).setUp()

        self.mock_agent = mock.MagicMock()
        self.handler = agent_base.CNAEventHandler(self.mock_agent)

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'CNAEventHandler._prov_reqs_for_uri')
    def test_process(self, mock_prov):
        events = {'URI1': 'add', 'URI2': 'delete', 'URI3': 'invalidate'}
        self.handler.process(events)

        # URI2 shouldn't be invoked.
        self.assertEqual(2, mock_prov.call_count)
        mock_prov.assert_any_call('URI1')
        mock_prov.assert_any_call('URI3')

    def test_prov_reqs_for_uri_not_lpar(self):
        """Ensures that anything but a LogicalPartition returns empty."""
        vio_uri = ('https://9.1.2.3:12443/rest/api/uom/ManagedSystem/'
                   'c5d782c7-44e4-3086-ad15-b16fb039d63b/VirtualIOServer/'
                   '3443DB77-AED1-47ED-9AA5-3DB9C6CF7089')
        self.assertEqual([], self.handler._prov_reqs_for_uri(vio_uri))

        ms_uri = ('https://9.1.2.3:12443/rest/api/uom/ManagedSystem/'
                  'c5d782c7-44e4-3086-ad15-b16fb039d63b')
        self.assertEqual([], self.handler._prov_reqs_for_uri(ms_uri))

        bad_uri = ('https://9.1.2.3')
        self.assertEqual([], self.handler._prov_reqs_for_uri(bad_uri))

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.list_cnas')
    def test_prov_reqs_for_uri(self, mock_list_cnas):
        """Happy path testing of prov_reqs_for_uri."""
        lpar_uri = ('https://9.1.2.3:12443/rest/api/uom/ManagedSystem/'
                    'c5d782c7-44e4-3086-ad15-b16fb039d63b/LogicalPartition/'
                    '3443DB77-AED1-47ED-9AA5-3DB9C6CF7089')

        cna1 = mock.MagicMock(mac='aabbccddeeff')
        cna2 = mock.MagicMock(mac='aabbccddee11')
        mock_list_cnas.return_value = [cna1, cna2]

        resp = self.handler._prov_reqs_for_uri(lpar_uri)

        self.assertEqual(2, len(resp))
        for p_req in resp:
            self.assertIsInstance(p_req, agent_base.ProvisionRequest)

        # Called the correct macs with the CNA.
        self.mock_agent.get_device_details.assert_any_call('aa:bb:cc:dd:ee:ff')
        self.mock_agent.get_device_details.assert_any_call('aa:bb:cc:dd:ee:11')
